#!/bin/ash
# author  : Xing Qingjie
# version : 0.03
# history : 12/15/2022  v0.03 adapt to 2x02 equipment
#           06/08/2021	v0.02 kernel 4.9-5.2 meminfo doesn't display VmallocUsed
#           03/14/2021  v0.01 init draft

# socket queue Recv-Q and Send-Q limit(Bytes)
MEM_THRESHOLD=80
# socket queue Recv-Q and Send-Q limit(Bytes)
SOCK_QUEUE_LIMIT=500000
# black hole memory limit(KBytes)
BLACK_HOLE_LIMIT=500000
# /tmp directory memory limit(KBytes)
TMP_MEM_LIMIT=100000

# color print functions
echoGreen()
{
    content=$@
    echo -e "\033[32m${content}\033[0m"
}

echoYellow()
{
    content=$@
    echo -e "\033[33m${content}\033[0m"
}

echoRed()
{
    content=$@
    echo -e "\033[31m${content}\033[0m"
}

echoBlue()
{
    content=$@
    echo -e "\033[34m${content}\033[0m"
}

# parse /proc/meminfo and set the relative parameters
get_meminfo()
{
    MemTotal=`cat /proc/meminfo | grep -w MemTotal | awk '{print $2}'`
    MemFree=`cat /proc/meminfo | grep -w MemFree | awk '{print $2}'`
    MemAvailable=`cat /proc/meminfo | grep -w MemAvailable | awk '{print $2}'`
    Buffers=`cat /proc/meminfo | grep -w Buffers | awk '{print $2}'`
    Cached=`cat /proc/meminfo | grep -w Cached | awk '{print $2}'`

    Active=`cat /proc/meminfo | grep -w Active: | awk '{print $2}'`
    Inactive=`cat /proc/meminfo | grep -w Inactive: | awk '{print $2}'`

    Unevictable=`cat /proc/meminfo | grep -w Unevictable | awk '{print $2}'`

    Slab=`cat /proc/meminfo | grep -w Slab | awk '{print $2}'`

    KernelStack=`cat /proc/meminfo | grep -w KernelStack | awk '{print $2}'`
    PageTables=`cat /proc/meminfo | grep -w PageTables | awk '{print $2}'`
    Bounce=`cat /proc/meminfo | grep -w Bounce | awk '{print $2}'`

    VmallocUsed=`cat /proc/meminfo | grep -w VmallocUsed | awk '{print $2}'`

    HardwareCorrupted=`cat /proc/meminfo | grep -w HardwareCorrupted | awk '{print $2}'`

    HugePages_Total=`cat /proc/meminfo | grep -w HugePages_Total | awk '{print $2}'`
    Hugepagesize=`cat /proc/meminfo | grep -w Hugepagesize | awk '{print $2}'`

    HardwareCorrupted=`cat /proc/meminfo | grep -w HardwareCorrupted | awk '{print $2}'`
}

diagnose_socket_queue()
{
    # check socket Recv-Q and Send-Q
    overlimit=false

    old=$IFS
    IFS=$'\n'
    for i in `netstat -aupt | sed 1,2d`; do
       local rcv=`echo $i | awk '{print $2}'`
       local snd=`echo $i | awk '{print $3}'`
       if [ $rcv -gt $SOCK_QUEUE_LIMIT ]; then
           # process Recv-Q 
           overlimit=true
           echoRed socket Recv-Q over limited
           echo $i 
       fi

       if [ $snd -gt $SOCK_QUEUE_LIMIT ]; then
           # process Send-Q 
           overlimit=true
           echoRed socket Send-Q over limited
           echo $i 
       fi
    done
    IFS=$old
}

diagnose_black_mem()
{
    # check virutal machine
    echoYellow ======== check black hole memory ========
    mod_name=`lsmod | grep balloon | awk '{print $1}'`
    if [ -n "$mod_name" ]; then
        # virtual machine with balloon tech
        echoRed virtual machine with $mod_name module
        rmmod $mod_name
    fi

    # check socket queue
    diagnose_socket_queue

}

diagnose_user_mem()
{
    # check /tmp memory
    echoYellow ======== check user space memory ========

    tmp_mem=`du -s /tmp/ | awk '{printf"%d",$1}'` 

    if [ $tmp_mem -gt $TMP_MEM_LIMIT ]; then
        echoRed tmp directory occupies too much memory $tmp_mem. please clean it
    else
        echoRed please check these proesses:
        old=$IFS
        IFS=$'\n'
        top_list=`top -n 1 | head -10 | sed 1,3d`
        echo "${top_list}"
        IFS=$old
    fi 
}

diagnose_kernel_mem()
{
    echoYellow ======== check kernel space memory ========

    # check slab memory
    slab_usage=`echo "$Slab $MemTotal" | awk '{printf("%.2f",$1/$2)}'`
    slab_usage_percent=`echo "$slab_usage" | awk '{printf("%.2f",$1*100)}'`
    
    a=3
    b=8
    MEM_THRESHOLD_tmp1=`echo "$a $b" | awk '{printf("%.2f",$1/$2)}'`
    MEM_THRESHOLD_tmp2=`echo "$MEM_THRESHOLD $MEM_THRESHOLD_tmp1" | awk '{printf("%.2f",$1*$2)}'`
    MEM_THRESHOLD_tmp3=`echo "$slab_usage_percent $MEM_THRESHOLD_tmp2" | awk '{if($1>$2) {printf"%d",1} else {printf"%d",0}}'`
    
    if [ $MEM_THRESHOLD_tmp3 -eq 1 ]; then
        echoRed slab memory is suspicious, please check
        slab_list=`cat /proc/slabinfo`
        echo "${slab_list}"
    else
        echoRed vmalloc memory is suspicious, please check
        old=$IFS
        IFS=$'\n'
        vmallc_list=`cat /proc/vmallocinfo`
        for i in $vmallc_list; do
            type=`echo $i | awk '{print $3}'`
            vmem=`echo $i | awk '{print $2}'`
            echo "$type : $vmem" >> vmallc.txt
        done
        IFS=$old

        vmallc_list=`sort vmallc.txt`
        rm -f vmallc.txt
        echo "$vmallc_list" >> vmallc.txt

        key=`echo $vmallc_list | awk '{print $1}'`
        vmem=0
        while read line; do
          type_tmp=`echo $line | awk '{print $1}'`
          vmem_tmp=`echo $line | awk '{print $3}'`
           if [ $key = $type_tmp ];then
              vmem=`echo $vmem $vmem_tmp | awk '{printf("%d", $1+$2)}'`
           else
              echo "$key : $vmem"
              key=$type_tmp
              vmem=$vmem_tmp
           fi
        done < vmallc.txt
        rm vmallc.txt
    fi

}

diagnose_huge_mem()
{
    echoYellow ======== check huge page memory ========
    echoRed HugePage consumption $hugeMem HugePages_Total=$HugePages_Total Hugepagesize=$Hugepagesize
}

#### main process ####

get_meminfo

if [ $VmallocUsed -eq 0 ]; then
    old=$IFS
    IFS=$'\n'
    vmallc_list=`cat /proc/vmallocinfo`
    for i in $vmallc_list; do
        vmem=`echo $i | awk '{print $2}'`
        VmallocUsed=$(($VmallocUsed+$vmem))
    done
    IFS=$old
fi

usedMem=`echo " $MemTotal $MemFree" | awk '{printf("%.2f",$1-$2)}'`
kernelMem=`echo " $Slab $MemFree $PageTables $KernelStack $HardwareCorrupted $Bounce" | awk '{printf("%.2f",$1+$2+$3+$4+$5+$6)}'`
userMem=`echo " $Active $Inactive $Unevictable" | awk '{printf("%.2f",$1+$2+$3)}'`
hugeMem=`echo " $HugePages_Total $Hugepagesize" | awk '{printf("%.2f",$1*$2)}'`

## check memory usage
tmp1=`echo "$MemAvailable $MemTotal" | awk '{printf("%.2f",$1/$2)}'`
tmp2=`echo "$tmp1" | awk '{printf("%.2f",1-$1)}'`
mem_usage=`echo "$tmp2" | awk '{printf("%.2f",$1*100)}'`
mem_usage_percent=$mem_usage

tmp=`echo "$mem_usage_percent $MEM_THRESHOLD" | awk '{if($1<=$2) {printf"%d",1} else {printf"%d",0}}'`
if [ $tmp -eq 1 ];then
    echoGreen memory usages $mem_usage_percent%, no need to diagnose
    exit 0
fi

## check black hole memory
blackmem=`echo " $usedMem $userMem $kernelMem $hugeMem" | awk '{printf("%d",$1-$2-$3-$4)}'`
user_usage_percent_tmp=`echo "$userMem $MemTotal" | awk '{printf("%.2f",$1/$2)}'`
user_usage_percent=`echo "$user_usage_percent_tmp" | awk '{printf("%.2f",$1*100)}'`
kernel_usage_percent_tmp=`echo "$kernelMem $MemTotal" | awk '{printf("%.2f",$1/$2)}'`
kernel_usage_percent=`echo "$kernel_usage_percent_tmp" | awk '{printf("%.2f",$1*100)}'`
huge_usage_percent_tmp=`echo "$hugeMem $MemTotal" | awk '{printf("%.2f",$1/$2)}'`
huge_usage_percent=`echo "$huge_usage_percent_tmp" | awk '{printf("%.2f",$1*100)}'`

MEM_THRESHOLD_tmp4=`echo "$MEM_THRESHOLD" | awk '{printf("%.2f",$1/2)}'`

if [ $blackmem -gt $BLACK_HOLE_LIMIT ]; then
    diagnose_black_mem
elif [ `echo "$huge_usage_percent $MEM_THRESHOLD_tmp4" | awk '{if($1>$2) {printf"%d",1} else {printf"%d",0}}'` -eq 1 ]; then
    diagnose_huge_mem
elif [ `echo "$user_usage_percent $MEM_THRESHOLD_tmp4" | awk '{if($1>$2) {printf"%d",1} else {printf"%d",0}}'` -eq 1 ]; then
    diagnose_user_mem
else
    diagnose_kernel_mem
fi
